import math
from typing import Optional

import numpy as np
from tomolab.Reconstruction.SPECT import SPECT_Static_Scan

from mu_map.dataset.util import align_images

COLOR_BLACK = (0, 0, 0)
COLOR_WHITE = (255, 255, 255)


def to_grayscale(
    img: np.ndarray, min_val: Optional[float] = None, max_val: Optional[float] = None
):
    """
    Convert an arbitrary image to a grayscale image with a value range of [0, 255].

    :param img: the image to be converted to grayscale
    :param min_val: minimum value used for normalization, this is helpful if the image has to be normalized relative to others
    :param max_val: maximum value used for normalization, this is helpful if the image has to be normalized relative to others
    :return: the image in grayscale
    """
    if min_val is None:
        min_val = img.min()

    if max_val is None:
        max_val = img.max()

    if (max_val - min_val) == 0:
        return np.zeros(img.shape, np.uint8)

    _img = (img - min_val) / (max_val - min_val)
    _img = (_img * 255).astype(np.uint8)
    return _img


def grayscale_to_rgb(img: np.ndarray):
    """
    Convert a grayscale image to an rgb image by repeating it three times.

    :param img: the grayscale image to be converted to rgb
    :return: the image in rgb
    """
    assert img.ndim == 2, f"grascale image has more than 2 dimensions {img.shape}"
    return img.repeat(3).reshape((*img.shape, 3))


import random
def reconstruct(recon, mu_map=None, use_gpu=True, seed=42):
    random.seed(42)

    spect = SPECT_Static_Scan()
    spect.set_use_gpu(use_gpu)

    recon_t = np.transpose(recon, (2, 1, 0)).astype(np.float32)
    n_pixels = recon_t.shape[0]
    print(f"NPixels {n_pixels}, {recon_t.shape}")
    # spect.set_n_pixels(n_pixels, n_pixels)
    spect.set_n_pixels(128, 128)
    spect.set_gantry_angular_positions(0.0, 360.0, 59)

    if mu_map is not None:
        mu_map_t = np.transpose(mu_map, (2, 1, 0)).astype(np.float32)

        padding_lower = math.ceil(n_pixels - mu_map_t.shape[2])
        padding_upper = math.floor(n_pixels - mu_map_t.shape[2])
        mu_map_t = np.pad(mu_map_t, [(0, 0), (0, 0), (padding_lower, padding_upper)])

        spect.set_attenuation(mu_map_t)

    print("Forward projection .....", end="\r")
    measurement = spect.project(recon_t)
    print(f"Measurement: {measurement.data.shape}")
    spect.set_measurement(measurement.data)
    print("Forward projection DONE!")

    spect.set_pixel_size(4.8, 4.8)
    spect.set_radius(200.0)
    spect.set_psf(fwhm0_mm=5.0, depth_dependence=0.0001)

    print("Reconstruction .....", end="\r")
    activity = spect.estimate_activity(
        iterations=10, subset_size=16, subset_mode="random", method="EM"
    )
    print("Reconstruction DONE!")

    activity = activity.data
    activity = np.transpose(activity, (2, 1, 0))
    activity, _ = align_images(activity, recon)
    return activity

if __name__ == "__main__":
    import pydicom
    
    from mu_map.dataset.util import load_dcm_img

    img = load_dcm_img("./data/second/images/0001-stress-recon_nac_nsc.dcm")
    mu_map = load_dcm_img("./data/second/images/0001-stress-mu_map.dcm")

    reconstruct(img, mu_map)
